import json

MORPHO_AAVE_V3_COMPTROLLER_ABI = json.loads('''
[
  {
    "inputs": [],
    "name": "AddressIsZero",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "AmountIsZero",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnsafeCast",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "AssetIsCollateralOnMorpho",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "AssetNotCollateralOnMorpho",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "AssetNotCollateralOnPool",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BorrowIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BorrowNotEnabled",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BorrowNotPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ClaimRewardsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "CollateralIsZero",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "DebtIsZero",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ExceedsBorrowCap",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ExceedsMaxBasisPoints",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InconsistentEMode",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidNonce",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidSignatory",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidValueS",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidValueV",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "LiquidateBorrowIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "LiquidateCollateralIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "MarketAlreadyCreated",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "MarketIsDeprecated",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "MarketIsNotListedOnAave",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "MarketLtTooLow",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "MarketNotCreated",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "PermissionDenied",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "RepayIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SentinelBorrowNotEnabled",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SentinelLiquidateNotEnabled",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SetAsCollateralOnPoolButMarketNotCreated",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SignatureExpired",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SiloedBorrowMarket",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SupplyCollateralIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SupplyIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SupplyIsZero",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnauthorizedBorrow",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnauthorizedLiquidate",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnauthorizedWithdraw",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "WithdrawCollateralIsPaused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "WithdrawIsPaused",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferStarted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledOnPool",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledInP2P",
        "type": "uint256"
      }
    ],
    "name": "BorrowPositionUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "caller",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledOnPool",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledInP2P",
        "type": "uint256"
      }
    ],
    "name": "Borrowed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledBalance",
        "type": "uint256"
      }
    ],
    "name": "CollateralSupplied",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "caller",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledBalance",
        "type": "uint256"
      }
    ],
    "name": "CollateralWithdrawn",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint128",
        "name": "repay",
        "type": "uint128"
      },
      {
        "indexed": false,
        "internalType": "uint128",
        "name": "withdraw",
        "type": "uint128"
      }
    ],
    "name": "DefaultIterationsSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "idleSupply",
        "type": "uint256"
      }
    ],
    "name": "IdleSupplyUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "poolSupplyIndex",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "p2pSupplyIndex",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "poolBorrowIndex",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "p2pBorrowIndex",
        "type": "uint256"
      }
    ],
    "name": "IndexesUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsBorrowPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsClaimRewardsPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isCollateral",
        "type": "bool"
      }
    ],
    "name": "IsCollateralSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isDeprecated",
        "type": "bool"
      }
    ],
    "name": "IsDeprecatedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsLiquidateBorrowPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsLiquidateCollateralPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isP2PDisabled",
        "type": "bool"
      }
    ],
    "name": "IsP2PDisabledSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsRepayPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsSupplyCollateralPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsSupplyPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsWithdrawCollateralPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "IsWithdrawPausedSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "liquidator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlyingBorrowed",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountLiquidated",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "underlyingCollateral",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountSeized",
        "type": "uint256"
      }
    ],
    "name": "Liquidated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "delegator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "manager",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "isAllowed",
        "type": "bool"
      }
    ],
    "name": "ManagerApproval",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      }
    ],
    "name": "MarketCreated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledDelta",
        "type": "uint256"
      }
    ],
    "name": "P2PBorrowDeltaUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "P2PDeltasIncreased",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint16",
        "name": "p2pIndexCursor",
        "type": "uint16"
      }
    ],
    "name": "P2PIndexCursorSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledDelta",
        "type": "uint256"
      }
    ],
    "name": "P2PSupplyDeltaUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledTotalSupplyP2P",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledTotalBorrowP2P",
        "type": "uint256"
      }
    ],
    "name": "P2PTotalsUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "positionsManager",
        "type": "address"
      }
    ],
    "name": "PositionsManagerSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "repayer",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledOnPool",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledInP2P",
        "type": "uint256"
      }
    ],
    "name": "Repaid",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint16",
        "name": "reserveFactor",
        "type": "uint16"
      }
    ],
    "name": "ReserveFactorSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "claimed",
        "type": "uint256"
      }
    ],
    "name": "ReserveFeeClaimed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "claimer",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "rewardToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountClaimed",
        "type": "uint256"
      }
    ],
    "name": "RewardsClaimed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "rewardsManager",
        "type": "address"
      }
    ],
    "name": "RewardsManagerSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledOnPool",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledInP2P",
        "type": "uint256"
      }
    ],
    "name": "Supplied",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledOnPool",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledInP2P",
        "type": "uint256"
      }
    ],
    "name": "SupplyPositionUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "treasuryVault",
        "type": "address"
      }
    ],
    "name": "TreasuryVaultSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "caller",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "signatory",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "usedNonce",
        "type": "uint256"
      }
    ],
    "name": "UserNonceIncremented",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "caller",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledOnPool",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "scaledInP2P",
        "type": "uint256"
      }
    ],
    "name": "Withdrawn",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "DOMAIN_SEPARATOR",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "acceptOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "addressesProvider",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isAllowed",
        "type": "bool"
      }
    ],
    "name": "approveManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "delegator",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isAllowed",
        "type": "bool"
      },
      {
        "internalType": "uint256",
        "name": "nonce",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "deadline",
        "type": "uint256"
      },
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "v",
            "type": "uint8"
          },
          {
            "internalType": "bytes32",
            "name": "r",
            "type": "bytes32"
          },
          {
            "internalType": "bytes32",
            "name": "s",
            "type": "bytes32"
          }
        ],
        "internalType": "struct Types.Signature",
        "name": "signature",
        "type": "tuple"
      }
    ],
    "name": "approveManagerWithSig",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "maxIterations",
        "type": "uint256"
      }
    ],
    "name": "borrow",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "borrowBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      }
    ],
    "name": "claimRewards",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "rewardTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "claimedAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "underlyings",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "amounts",
        "type": "uint256[]"
      }
    ],
    "name": "claimToTreasury",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "collateralBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint16",
        "name": "reserveFactor",
        "type": "uint16"
      },
      {
        "internalType": "uint16",
        "name": "p2pIndexCursor",
        "type": "uint16"
      }
    ],
    "name": "createMarket",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "defaultIterations",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint128",
            "name": "repay",
            "type": "uint128"
          },
          {
            "internalType": "uint128",
            "name": "withdraw",
            "type": "uint128"
          }
        ],
        "internalType": "struct Types.Iterations",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "eModeCategoryId",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "enum Types.Position",
        "name": "position",
        "type": "uint8"
      }
    ],
    "name": "getBucketsMask",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "enum Types.Position",
        "name": "position",
        "type": "uint8"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "getNext",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "increaseP2PDeltas",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "addressesProvider",
        "type": "address"
      },
      {
        "internalType": "uint8",
        "name": "eModeCategoryId",
        "type": "uint8"
      },
      {
        "internalType": "address",
        "name": "positionsManager",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "uint128",
            "name": "repay",
            "type": "uint128"
          },
          {
            "internalType": "uint128",
            "name": "withdraw",
            "type": "uint128"
          }
        ],
        "internalType": "struct Types.Iterations",
        "name": "defaultIterations",
        "type": "tuple"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isClaimRewardsPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "delegator",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      }
    ],
    "name": "isManagedBy",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlyingBorrowed",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "underlyingCollateral",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "liquidate",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "liquidityData",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "borrowable",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "maxDebt",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "debt",
            "type": "uint256"
          }
        ],
        "internalType": "struct Types.LiquidityData",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      }
    ],
    "name": "market",
    "outputs": [
      {
        "components": [
          {
            "components": [
              {
                "components": [
                  {
                    "internalType": "uint128",
                    "name": "poolIndex",
                    "type": "uint128"
                  },
                  {
                    "internalType": "uint128",
                    "name": "p2pIndex",
                    "type": "uint128"
                  }
                ],
                "internalType": "struct Types.MarketSideIndexes",
                "name": "supply",
                "type": "tuple"
              },
              {
                "components": [
                  {
                    "internalType": "uint128",
                    "name": "poolIndex",
                    "type": "uint128"
                  },
                  {
                    "internalType": "uint128",
                    "name": "p2pIndex",
                    "type": "uint128"
                  }
                ],
                "internalType": "struct Types.MarketSideIndexes",
                "name": "borrow",
                "type": "tuple"
              }
            ],
            "internalType": "struct Types.Indexes",
            "name": "indexes",
            "type": "tuple"
          },
          {
            "components": [
              {
                "components": [
                  {
                    "internalType": "uint256",
                    "name": "scaledDelta",
                    "type": "uint256"
                  },
                  {
                    "internalType": "uint256",
                    "name": "scaledP2PTotal",
                    "type": "uint256"
                  }
                ],
                "internalType": "struct Types.MarketSideDelta",
                "name": "supply",
                "type": "tuple"
              },
              {
                "components": [
                  {
                    "internalType": "uint256",
                    "name": "scaledDelta",
                    "type": "uint256"
                  },
                  {
                    "internalType": "uint256",
                    "name": "scaledP2PTotal",
                    "type": "uint256"
                  }
                ],
                "internalType": "struct Types.MarketSideDelta",
                "name": "borrow",
                "type": "tuple"
              }
            ],
            "internalType": "struct Types.Deltas",
            "name": "deltas",
            "type": "tuple"
          },
          {
            "internalType": "address",
            "name": "underlying",
            "type": "address"
          },
          {
            "components": [
              {
                "internalType": "bool",
                "name": "isP2PDisabled",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isSupplyPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isSupplyCollateralPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isBorrowPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isWithdrawPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isWithdrawCollateralPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isRepayPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isLiquidateCollateralPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isLiquidateBorrowPaused",
                "type": "bool"
              },
              {
                "internalType": "bool",
                "name": "isDeprecated",
                "type": "bool"
              }
            ],
            "internalType": "struct Types.PauseStatuses",
            "name": "pauseStatuses",
            "type": "tuple"
          },
          {
            "internalType": "bool",
            "name": "isCollateral",
            "type": "bool"
          },
          {
            "internalType": "address",
            "name": "variableDebtToken",
            "type": "address"
          },
          {
            "internalType": "uint32",
            "name": "lastUpdateTimestamp",
            "type": "uint32"
          },
          {
            "internalType": "uint16",
            "name": "reserveFactor",
            "type": "uint16"
          },
          {
            "internalType": "uint16",
            "name": "p2pIndexCursor",
            "type": "uint16"
          },
          {
            "internalType": "address",
            "name": "aToken",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "stableDebtToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "idleSupply",
            "type": "uint256"
          }
        ],
        "internalType": "struct Types.Market",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "marketsCreated",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pendingOwner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pool",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "positionsManager",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      }
    ],
    "name": "repay",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "deadline",
        "type": "uint256"
      },
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "v",
            "type": "uint8"
          },
          {
            "internalType": "bytes32",
            "name": "r",
            "type": "bytes32"
          },
          {
            "internalType": "bytes32",
            "name": "s",
            "type": "bytes32"
          }
        ],
        "internalType": "struct Types.Signature",
        "name": "signature",
        "type": "tuple"
      }
    ],
    "name": "repayWithPermit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "rewardsManager",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "scaledCollateralBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "scaledP2PBorrowBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "scaledP2PSupplyBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "scaledPoolBorrowBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "scaledPoolSupplyBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isCollateral",
        "type": "bool"
      }
    ],
    "name": "setAssetIsCollateral",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isCollateral",
        "type": "bool"
      }
    ],
    "name": "setAssetIsCollateralOnPool",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "uint128",
            "name": "repay",
            "type": "uint128"
          },
          {
            "internalType": "uint128",
            "name": "withdraw",
            "type": "uint128"
          }
        ],
        "internalType": "struct Types.Iterations",
        "name": "defaultIterations",
        "type": "tuple"
      }
    ],
    "name": "setDefaultIterations",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsBorrowPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsClaimRewardsPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isDeprecated",
        "type": "bool"
      }
    ],
    "name": "setIsDeprecated",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsLiquidateBorrowPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsLiquidateCollateralPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isP2PDisabled",
        "type": "bool"
      }
    ],
    "name": "setIsP2PDisabled",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsPausedForAllMarkets",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsRepayPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsSupplyCollateralPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsSupplyPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsWithdrawCollateralPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isPaused",
        "type": "bool"
      }
    ],
    "name": "setIsWithdrawPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint16",
        "name": "p2pIndexCursor",
        "type": "uint16"
      }
    ],
    "name": "setP2PIndexCursor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "positionsManager",
        "type": "address"
      }
    ],
    "name": "setPositionsManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint16",
        "name": "newReserveFactor",
        "type": "uint16"
      }
    ],
    "name": "setReserveFactor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "rewardsManager",
        "type": "address"
      }
    ],
    "name": "setRewardsManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "treasuryVault",
        "type": "address"
      }
    ],
    "name": "setTreasuryVault",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "maxIterations",
        "type": "uint256"
      }
    ],
    "name": "supply",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "supplyBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      }
    ],
    "name": "supplyCollateral",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "deadline",
        "type": "uint256"
      },
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "v",
            "type": "uint8"
          },
          {
            "internalType": "bytes32",
            "name": "r",
            "type": "bytes32"
          },
          {
            "internalType": "bytes32",
            "name": "s",
            "type": "bytes32"
          }
        ],
        "internalType": "struct Types.Signature",
        "name": "signature",
        "type": "tuple"
      }
    ],
    "name": "supplyCollateralWithPermit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "maxIterations",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "deadline",
        "type": "uint256"
      },
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "v",
            "type": "uint8"
          },
          {
            "internalType": "bytes32",
            "name": "r",
            "type": "bytes32"
          },
          {
            "internalType": "bytes32",
            "name": "s",
            "type": "bytes32"
          }
        ],
        "internalType": "struct Types.Signature",
        "name": "signature",
        "type": "tuple"
      }
    ],
    "name": "supplyWithPermit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "treasuryVault",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      }
    ],
    "name": "updatedIndexes",
    "outputs": [
      {
        "components": [
          {
            "components": [
              {
                "internalType": "uint256",
                "name": "poolIndex",
                "type": "uint256"
              },
              {
                "internalType": "uint256",
                "name": "p2pIndex",
                "type": "uint256"
              }
            ],
            "internalType": "struct Types.MarketSideIndexes256",
            "name": "supply",
            "type": "tuple"
          },
          {
            "components": [
              {
                "internalType": "uint256",
                "name": "poolIndex",
                "type": "uint256"
              },
              {
                "internalType": "uint256",
                "name": "p2pIndex",
                "type": "uint256"
              }
            ],
            "internalType": "struct Types.MarketSideIndexes256",
            "name": "borrow",
            "type": "tuple"
          }
        ],
        "internalType": "struct Types.Indexes256",
        "name": "indexes",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "userBorrows",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "userCollaterals",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "userNonce",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "maxIterations",
        "type": "uint256"
      }
    ],
    "name": "withdraw",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "underlying",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "onBehalf",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      }
    ],
    "name": "withdrawCollateral",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


import json

REWARD_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "governor_",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "AlreadyConfigured",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "InvalidUInt64",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "NotPermitted",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "NotSupported",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "TransferOutFailed",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "oldGovernor",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newGovernor",
        "type": "address"
      }
    ],
    "name": "GovernorTransferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "RewardClaimed",
    "type": "event"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comet",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "shouldAccrue",
        "type": "bool"
      }
    ],
    "name": "claim",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comet",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "shouldAccrue",
        "type": "bool"
      }
    ],
    "name": "claimTo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comet",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getRewardOwed",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "token",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "owed",
            "type": "uint256"
          }
        ],
        "internalType": "struct CometRewards.RewardOwed",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "governor",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardConfig",
    "outputs": [
      {
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "internalType": "uint64",
        "name": "rescaleFactor",
        "type": "uint64"
      },
      {
        "internalType": "bool",
        "name": "shouldUpscale",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardsClaimed",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comet",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "token",
        "type": "address"
      }
    ],
    "name": "setRewardConfig",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newGovernor",
        "type": "address"
      }
    ],
    "name": "transferGovernor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "withdrawToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


import itertools

from query_state_lib.base.mappers.eth_call_balance_of_mapper import EthCallBalanceOf
from query_state_lib.base.mappers.get_balance_mapper import GetBalance
from query_state_lib.client.client_querier import ClientQuerier
from web3 import Web3, HTTPProvider
from web3.middleware import geth_poa_middleware

from defi_services.abis.dex.pancakeswap.pancakeswap_lp_token_abi import LP_TOKEN_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.abis.token.erc721_abi import ERC721_ABI
from defi_services.constants.network_constants import NATIVE_TOKEN
from defi_services.services.eth.eth_services import EthService
from defi_services.services.multicall.batch_queries_service import add_rpc_call, decode_data_response_ignore_error
from defi_services.utils.logger_utils import get_logger

logger = get_logger('State Query Service')


class StateQueryService:
    def __init__(self, provider_uri):
        self._w3 = Web3(HTTPProvider(provider_uri))
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        self.client_querier = ClientQuerier(provider_url=provider_uri)

    def get_web3(self):
        return self._w3

    def to_checksum(self, address):
        return self._w3.to_checksum_address(address)

    def get_contract(self, address, abi):
        try:
            return self._w3.eth.contract(self.to_checksum(address), abi=abi)
        except TypeError:
            print('a')

    def is_contract(self, address):
        code = self._w3.eth.get_code(self.to_checksum(address))
        code_str = code.hex()
        if code_str == '0x':
            return False
        else:
            return True

    def decode_tx(self, address, abi, tx_input):
        contract = self.get_contract(address, abi)
        func_obj, func_params = contract.decode_function_input(tx_input)
        return func_obj, func_params

    def is_address(self, address):
        return self._w3.is_address(address)

    def balance_of(self, address, token, abi=ERC20_ABI, block_number='latest'):
        token_contract = self._w3.eth.contract(self._w3.to_checksum_address(token), abi=abi)
        decimals = token_contract.functions.decimals().call()
        balance = token_contract.functions.balanceOf(self._w3.to_checksum_address(address)).call(
            block_identifier=block_number)
        return balance / 10 ** decimals

    def batch_balance_of(self, address, tokens, block_number: int = 'latest', batch_size=100):
        balances = {}
        tokens = {token['address']: token['decimals'] or 18 for token in tokens}
        rpc_requests = []
        for token in tokens:
            if token != "0x" and token != NATIVE_TOKEN:
                query_id = f'{address}_{token}'
                call_balance_of = EthCallBalanceOf(
                    contract_address=Web3.to_checksum_address(token),
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
            else:
                query_id = f"{address}_{NATIVE_TOKEN}"
                call_balance_of = GetBalance(
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
            rpc_requests.append(call_balance_of)
        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)
        for token, decimals in tokens.items():
            if token == '0x':
                token = NATIVE_TOKEN
            balance = rpc_responses.get(f'{address}_{token}').result or 0
            balance = balance / 10 ** decimals
            balances[token] = balance

        return balances

    def batch_balance_of_multiple_addresses(self, addresses, tokens, block_number='latest', batch_size=100):
        tokens_decimals = {token['address']: token.get('decimals') or 18 for token in tokens}

        rpc_requests = []
        for address, interacted_tokens in addresses.items():
            for token_address in interacted_tokens:
                if token_address != "0x" and token_address != NATIVE_TOKEN:
                    query_id = f'{address}_{token_address}'
                    call_balance_of = EthCallBalanceOf(
                        contract_address=Web3.to_checksum_address(token_address),
                        address=Web3.to_checksum_address(address),
                        block_number=block_number,
                        id=query_id
                    )
                else:
                    query_id = f"{address}_{NATIVE_TOKEN}"
                    call_balance_of = GetBalance(
                        address=Web3.to_checksum_address(address),
                        block_number=block_number,
                        id=query_id
                    )
                rpc_requests.append(call_balance_of)

        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)

        results = {}
        for address, interacted_tokens in addresses.items():
            results[address] = {}
            for token_address in interacted_tokens:
                balance = rpc_responses.get(f'{address}_{token_address}').result or 0

                decimals = tokens_decimals.get(token_address) or 18
                balance = balance / 10 ** decimals
                results[address][token_address] = balance

        return results, len(rpc_requests)

    def get_block_number_by_timestamp(self, block_timestamp):
        eth_service = EthService(self._w3)
        block_number = eth_service.get_block_for_timestamp(block_timestamp)
        return block_number

    def batch_native_balance_of_wallets(self, addresses, blocks, decimals=18, batch_size=100):
        rpc_requests = []
        for address in addresses:
            for block_number in blocks:
                query_id = f"{address}_{block_number}"
                call_balance_of = GetBalance(
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
                rpc_requests.append(call_balance_of)

        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)

        data = {}
        for address in addresses:
            data[address] = {}
            for block_number in blocks:
                balance = rpc_responses.get(f'{address}_{block_number}').result or 0
                balance = balance / 10 ** decimals
                data[address][block_number] = balance

        return data

    def batch_balance_of_wallets(self, addresses, tokens, blocks, batch_size=100):
        rpc_requests = []

        for address, token, block_number in itertools.product(addresses, tokens, blocks):
            token_address = token['address']
            query_id = f'{address}_{token_address}_{block_number}'
            if token_address != "0x" and token_address != NATIVE_TOKEN:
                call_balance_of = EthCallBalanceOf(
                    contract_address=Web3.to_checksum_address(token_address),
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
            else:
                call_balance_of = GetBalance(
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
            rpc_requests.append(call_balance_of)

        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)

        data = {}
        for address, token, block_number in itertools.product(addresses, tokens, blocks):
            token_address = token['address']
            query_id = f'{address}_{token_address}_{block_number}'
            balance = rpc_responses.get(query_id).result or 0

            decimals = token.get('decimals') or 18

            if address not in data:
                data[address] = {}
            if token_address not in data[address]:
                data[address][token_address] = {}
            data[address][token_address][block_number] = balance / 10 ** decimals

        return data

    def batch_balance_of_wallets_block(self, addresses, tokens, block_number='latest', batch_size=100):
        rpc_requests = []
        for address, token in itertools.product(addresses, tokens):
            token_address = token['address']
            query_id = f'{address}_{token_address}_{block_number}'
            if token_address != "0x" and token_address != NATIVE_TOKEN:
                call_balance_of = EthCallBalanceOf(
                    contract_address=Web3.to_checksum_address(token_address),
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
            else:
                call_balance_of = GetBalance(
                    address=Web3.to_checksum_address(address),
                    block_number=block_number,
                    id=query_id
                )
            rpc_requests.append(call_balance_of)

        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)

        data = {}
        for address, token in itertools.product(addresses, tokens):
            token_address = token['address']
            query_id = f'{address}_{token_address}_{block_number}'
            balance = rpc_responses.get(query_id).result or 0
            decimals = token.get('decimals') or 18
            if address not in data:
                data[address] = {}
            if token_address not in data[address]:
                data[address][token_address] = {}
            data[address][token_address] = balance / 10 ** decimals

        return data

    def batch_balance_of_wallet_info(self, wallet_info: dict, block_number='latest', batch_size=100):
        rpc_requests = []
        for address, tokens in wallet_info.items():
            for token in tokens:
                token_address = token['address']
                query_id = f'{address}_{token_address}_{block_number}'
                if token_address != "0x" and token_address != NATIVE_TOKEN:
                    call_balance_of = EthCallBalanceOf(
                        contract_address=Web3.to_checksum_address(token_address),
                        address=Web3.to_checksum_address(address),
                        block_number=block_number,
                        id=query_id
                    )
                else:
                    call_balance_of = GetBalance(
                        address=Web3.to_checksum_address(address),
                        block_number=block_number,
                        id=query_id
                    )
                rpc_requests.append(call_balance_of)

        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)

        data = {}
        for address, tokens in wallet_info.items():
            for token in tokens:
                token_address = token['address']
                query_id = f'{address}_{token_address}_{block_number}'
                balance = rpc_responses.get(query_id).result or 0
                decimals = token.get('decimals') or 18
                if address not in data:
                    data[address] = {}
                if token_address not in data[address]:
                    data[address][token_address] = {}
                data[address][token_address] = balance / 10 ** decimals

        return data

    def batch_balance_of_wallet_info_with_block_number(self, wallets_info: list, batch_size=100):
        rpc_requests = []
        for wallet_info in wallets_info:
            address = wallet_info['address']
            tokens = wallet_info.get('tokens', [])
            block_number = wallet_info.get('block_number', 'latest')

            for token in tokens:
                token_address = token['address']
                query_id = f'{address}_{token_address}_{block_number}'
                if token_address != "0x" and token_address != NATIVE_TOKEN:
                    call_balance_of = EthCallBalanceOf(
                        contract_address=Web3.to_checksum_address(token_address),
                        address=Web3.to_checksum_address(address),
                        block_number=block_number,
                        id=query_id
                    )
                else:
                    call_balance_of = GetBalance(
                        address=Web3.to_checksum_address(address),
                        block_number=block_number,
                        id=query_id
                    )
                rpc_requests.append(call_balance_of)

        rpc_responses = self.client_querier.sent_batch_to_provider(rpc_requests, batch_size=batch_size)

        data = {}
        for wallet_info in wallets_info:
            address = wallet_info['address']
            tokens = wallet_info.get('tokens', [])
            block_number = wallet_info.get('block_number', 'latest')

            for token in tokens:
                token_address = token['address']
                query_id = f'{address}_{token_address}_{block_number}'
                balance = rpc_responses.get(query_id).result or 0
                decimals = token.get('decimals')
                if decimals is None:
                    continue

                if address not in data:
                    data[address] = {}
                if token_address not in data[address]:
                    data[address][token_address] = {}
                data[address][token_address] = balance / 10 ** decimals

        return data

    def recheck_liquidity_pool_v2_reserves(self, liquidity_pools: list, batch_size=100):
        list_rpc_call = []
        list_call_id = []

        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']
            block_number = liquidity_pool.get('block_number', 'latest')

            add_rpc_call(
                abi=LP_TOKEN_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="getReserves", block_number=block_number,
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )

        responses = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=batch_size)
        decoded_data = decode_data_response_ignore_error(data_responses=responses, list_call_id=list_call_id)

        data = {}
        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']
            tokens = liquidity_pool.get('tokens', [])
            block_number = liquidity_pool.get('block_number', 'latest')

            reserves = decoded_data.get(f'getReserves_{address}_{block_number}'.lower())
            if not reserves:
                continue
            for token in tokens:
                token_address = token['address']
                balance = reserves[token['idx']]
                decimals = token.get('decimals')
                if decimals is None:
                    continue

                if address not in data:
                    data[address] = {}
                if token_address not in data[address]:
                    data[address][token_address] = {}
                data[address][token_address] = balance / 10 ** decimals

        return data

    def batch_total_supply_pool_info_with_block_number(self, liquidity_pools: list, batch_size=100):
        list_rpc_call = []
        list_call_id = []
        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']
            block_number = liquidity_pool['block_number']

            add_rpc_call(
                abi=LP_TOKEN_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="totalSupply", block_number=block_number,
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )
            add_rpc_call(
                abi=LP_TOKEN_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="decimals", block_number="latest",
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )
        try:
            responses = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=batch_size)
            decoded_data = decode_data_response_ignore_error(data_responses=responses, list_call_id=list_call_id)
        except Exception as ex:
            err_detail = str(ex)
            if err_detail.strip().startswith('Response data err'):
                return
            raise ex
        data = {}
        for liquidity_pool in liquidity_pools:
            try:
                address = liquidity_pool['address']
                block_number = liquidity_pool['block_number']
                total_supply = decoded_data.get(f'totalSupply_{address}_{block_number}'.lower())
                decimals = decoded_data.get(f'decimals_{address}_latest'.lower())
                data[address] = {
                    "liquidityAmount": total_supply / 10 ** decimals,
                    'decimals': decimals,
                }

            except Exception:
                continue

        return data

    def batch_nfts_info(self, addresses, batch_size=100):
        list_rpc_call = []
        list_call_id = []
        for address in addresses:
            add_rpc_call(
                abi=ERC721_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="name", block_number='latest',
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )

            add_rpc_call(
                abi=ERC721_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="symbol", block_number='latest',
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )

        responses = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=batch_size)
        decoded_data = decode_data_response_ignore_error(data_responses=responses, list_call_id=list_call_id)

        data = {}
        for address in addresses:
            query_name_id = f'name_{address}_latest'
            query_symbol_id = f'symbol_{address}_latest'

            name = decoded_data.get(query_name_id)
            symbol = decoded_data.get(query_symbol_id)
            if name and symbol:
                data[address] = {
                    "name": name,
                    "symbol": symbol
                }

        return data

    def batch_liquidity_pools_tokens(self, liquidity_pools, batch_size=100):
        list_rpc_call = []
        list_call_id = []
        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']

            add_rpc_call(
                abi=LP_TOKEN_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="token0", block_number="latest",
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )
            add_rpc_call(
                abi=LP_TOKEN_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="token1", block_number="latest",
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )
            add_rpc_call(
                abi=LP_TOKEN_ABI, contract_address=Web3.to_checksum_address(address),
                fn_name="factory", block_number="latest",
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )

        try:
            responses = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=batch_size)
            decoded_data = decode_data_response_ignore_error(data_responses=responses, list_call_id=list_call_id)
        except Exception as ex:
            err_detail = str(ex)
            if err_detail.strip().startswith('Response data err'):
                return
            raise ex

        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']

            token0_address = decoded_data.get(f'token0_{address}_latest'.lower())
            token1_address = decoded_data.get(f'token1_{address}_latest'.lower())
            factory = decoded_data.get(f'factory_{address}_latest'.lower())

            if token0_address and token1_address:
                liquidity_pool.update({
                    "address": address,
                    "factoryAddress": factory,
                    "tokens": [
                        {
                            "idx": 0,
                            "address": token0_address.lower()
                        },
                        {
                            "idx": 1,
                            "address": token1_address.lower()
                        }
                    ]
                })

        return liquidity_pools

    def batch_liquidity_pools_info(self, liquidity_pools, batch_size=100):
        list_rpc_call = []
        list_call_id = []
        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']
            block_number = liquidity_pool['block_number']

            for token in liquidity_pool.get('tokens', []):
                add_rpc_call(
                    abi=ERC20_ABI, contract_address=Web3.to_checksum_address(token['address']),
                    fn_name="decimals", block_number='latest',
                    list_call_id=list_call_id, list_rpc_call=list_rpc_call
                )
                add_rpc_call(
                    abi=ERC20_ABI, contract_address=Web3.to_checksum_address(token['address']),
                    fn_name="symbol", block_number='latest',
                    list_call_id=list_call_id, list_rpc_call=list_rpc_call
                )
                add_rpc_call(
                    abi=ERC20_ABI, contract_address=Web3.to_checksum_address(token['address']),
                    fn_name="balanceOf", fn_paras=address, block_number=block_number,
                    list_call_id=list_call_id, list_rpc_call=list_rpc_call
                )

        try:
            responses = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size=batch_size)
            decoded_data = decode_data_response_ignore_error(data_responses=responses, list_call_id=list_call_id)
        except Exception as ex:
            err_detail = str(ex)
            if err_detail.strip().startswith('Response data err'):
                return
            raise ex

        for liquidity_pool in liquidity_pools:
            address = liquidity_pool['address']
            block_number = liquidity_pool['block_number']

            try:
                for token in liquidity_pool.get('tokens', []):
                    symbol = decoded_data.get(f'symbol_{token["address"]}_latest'.lower())
                    symbol = symbol.upper() if symbol else ""
                    decimals = decoded_data.get(f'decimals_{token["address"]}_latest'.lower())

                    liquidity_amount = decoded_data.get(f'balanceOf_{token["address"]}_{address}_{block_number}'.lower(), 0)
                    token['liquidityAmount'] = liquidity_amount / 10 ** decimals
                    token['symbol'] = symbol
                    token['decimals'] = decimals

                liquidity_pool['symbol'] = ' - '.join([t['symbol'] for t in liquidity_pool.get('tokens', [])])
            except Exception as ex:
                logger.exception(ex)

                # Add to invalid liquidity pools
                liquidity_pool.pop('tokens')

        return liquidity_pools



import json

LIDO_EXECUTION_LAYER_REWARDS_VAULT_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_lido",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_treasury",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "requestedBy",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "ERC20Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "requestedBy",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "ERC721Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "ETHReceived",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "LIDO",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "TREASURY",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "recoverERC20",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      }
    ],
    "name": "recoverERC721",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_maxAmount",
        "type": "uint256"
      }
    ],
    "name": "withdrawRewards",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "stateMutability": "payable",
    "type": "receive"
  }
]
''')


from defi_services.constants.time_constant import TimeConstants


def apr_to_apy(apr, seconds_per_year=TimeConstants.A_YEAR):
    apy = ((1 + (apr / seconds_per_year)) ** seconds_per_year) - 1
    return apy


def apy_to_apr(apy, seconds_per_year=TimeConstants.A_YEAR):
    apr = ((1 + apy) ** (1 / seconds_per_year) - 1) * seconds_per_year
    return apr



import json

COMET_EXT_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "bytes32",
            "name": "name32",
            "type": "bytes32"
          },
          {
            "internalType": "bytes32",
            "name": "symbol32",
            "type": "bytes32"
          }
        ],
        "internalType": "struct CometConfiguration.ExtConfiguration",
        "name": "config",
        "type": "tuple"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "BadAmount",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadNonce",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadSignatory",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidInt104",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidInt256",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidUInt104",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidUInt128",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidUInt64",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidValueS",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidValueV",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NegativeNumber",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SignatureExpired",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "Approval",
    "type": "event"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isAllowed_",
        "type": "bool"
      }
    ],
    "name": "allow",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "isAllowed_",
        "type": "bool"
      },
      {
        "internalType": "uint256",
        "name": "nonce",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "expiry",
        "type": "uint256"
      },
      {
        "internalType": "uint8",
        "name": "v",
        "type": "uint8"
      },
      {
        "internalType": "bytes32",
        "name": "r",
        "type": "bytes32"
      },
      {
        "internalType": "bytes32",
        "name": "s",
        "type": "bytes32"
      }
    ],
    "name": "allowBySig",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      }
    ],
    "name": "allowance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "approve",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseAccrualScale",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseIndexScale",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "baseTrackingAccrued",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "collateralBalanceOf",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "factorScale",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      }
    ],
    "name": "hasPermission",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "isAllowed",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "liquidatorPoints",
    "outputs": [
      {
        "internalType": "uint32",
        "name": "numAbsorbs",
        "type": "uint32"
      },
      {
        "internalType": "uint64",
        "name": "numAbsorbed",
        "type": "uint64"
      },
      {
        "internalType": "uint128",
        "name": "approxSpend",
        "type": "uint128"
      },
      {
        "internalType": "uint32",
        "name": "_reserved",
        "type": "uint32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "maxAssets",
    "outputs": [
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "name",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "priceScale",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "symbol",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalsBasic",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint64",
            "name": "baseSupplyIndex",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "baseBorrowIndex",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "trackingSupplyIndex",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "trackingBorrowIndex",
            "type": "uint64"
          },
          {
            "internalType": "uint104",
            "name": "totalSupplyBase",
            "type": "uint104"
          },
          {
            "internalType": "uint104",
            "name": "totalBorrowBase",
            "type": "uint104"
          },
          {
            "internalType": "uint40",
            "name": "lastAccrualTime",
            "type": "uint40"
          },
          {
            "internalType": "uint8",
            "name": "pauseFlags",
            "type": "uint8"
          }
        ],
        "internalType": "struct CometStorage.TotalsBasic",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "totalsCollateral",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "totalSupplyAsset",
        "type": "uint128"
      },
      {
        "internalType": "uint128",
        "name": "_reserved",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userBasic",
    "outputs": [
      {
        "internalType": "int104",
        "name": "principal",
        "type": "int104"
      },
      {
        "internalType": "uint64",
        "name": "baseTrackingIndex",
        "type": "uint64"
      },
      {
        "internalType": "uint64",
        "name": "baseTrackingAccrued",
        "type": "uint64"
      },
      {
        "internalType": "uint16",
        "name": "assetsIn",
        "type": "uint16"
      },
      {
        "internalType": "uint8",
        "name": "_reserved",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userCollateral",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "balance",
        "type": "uint128"
      },
      {
        "internalType": "uint128",
        "name": "_reserved",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userNonce",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "version",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')


import json

LIDO_DAO_INSURANCE_FUND_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_initialOwner",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "bytes",
        "name": "_data",
        "type": "bytes"
      }
    ],
    "name": "ERC1155Transferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ERC20Transferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "bytes",
        "name": "_data",
        "type": "bytes"
      }
    ],
    "name": "ERC721Transferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "EtherTransferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "_data",
        "type": "bytes"
      }
    ],
    "name": "transferERC1155",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "transferERC20",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "_data",
        "type": "bytes"
      }
    ],
    "name": "transferERC721",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "transferEther",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')